This file is a merged representation of the entire codebase, combining all repository files into a single document.
Generated by Repomix on: 2025-01-29T11:17:18.043Z

# File Summary

## Purpose
This file contains a packed representation of the entire repository's contents.
It is designed to be easily consumable by AI systems for analysis, code review,
or other automated processes.

## File Format
The content is organized as follows:
1. This summary section
2. Repository information
3. Directory structure
4. Multiple file entries, each consisting of:
  a. A header with the file path (## File: path/to/file)
  b. The full contents of the file in a code block

## Usage Guidelines
- This file should be treated as read-only. Any changes should be made to the
  original repository files, not this packed version.
- When processing this file, use the file path to distinguish
  between different files in the repository.
- Be aware that this file may contain sensitive information. Handle it with
  the same level of security as you would the original repository.

## Notes
- Some files may have been excluded based on .gitignore rules and Repomix's
  configuration.
- Binary files are not included in this packed representation. Please refer to
  the Repository Structure section for a complete list of file paths, including
  binary files.

## Additional Info

# Directory Structure
```
pyproject.toml
README.md
src/gp_diagnostics/__init__.py
src/gp_diagnostics/cv.py
src/gp_diagnostics/metrics.py
src/gp_diagnostics/plots.py
src/gp_diagnostics/utils/__init__.py
src/gp_diagnostics/utils/checks.py
src/gp_diagnostics/utils/linalg.py
src/gp_diagnostics/utils/stats.py
tests/test_cv.py
tests/test_metrics.py
tests/test_utils_checks.py
tests/test_utils_linalg.py
tests/test_utils_stats.py
tests/utils.py
tox.toml
```

# Files

## File: pyproject.toml
````toml
[build-system]
requires = ["hatchling"]
build-backend = "hatchling.build"

[project]
name = "gp-diagnostics"
version = "0.1.0"
description = "Tools for diagnosing Gaussian Process models"
readme = "README.md"
authors = [
    { name = "Christian Agrell", email = "christian.agrell@dnv.com" },
    { name = "Magnus Kristiansen", email = "magnus.kristiansen@dnv.com" },
]
requires-python = ">=3.10"
license = {file = "LICENSE"}
dependencies = [
    "numpy>=2.2.2",
    "pandas>=2.2.3",
    "plotly>=5.24.1",
    "scipy>=1.15.1",
]

[dependency-groups]
dev = [
    "gpytorch>=1.13",
    "jupyter>=1.1.1",
    "matplotlib>=3.10.0",
    "pytest>=8.3.4",
    "tox>=4.24.1",
    "tox-uv>=1.20.1",
]

[tool.ruff]
line-length = 120
exclude = ["examples/example.py", "tests/readme_example.py", "examples/exact_gp_example.ipynb", "examples/*"]  # TEMP

[tool.ruff.lint]
select = ["ALL"]
ignore = [
    "COM812",   # Missing trailing comma
    "ISC001",   # Single line implicit string concatination
    "N802",     # Function name should be lowercase
    "N803",     # Argument name should be lowercase
    "N806",     # Variable in function should be lowercase
    "PLR0912",  # Too many branches
    "PLR0913",  # Too many arguments in function definition
    "PLR0915",  # Too many statements
    "C901",     # Function is too complex
    "PT018",    # Assertion should be broken down into multiple parts
    "S101",     # Use of `assert` detected
    # TODO: fix the ones below, then remove them from the ignore list
    # "D",        # Docstrings
    # "ANN",      # Type annotations
    # "FBT002",   # Boolean default positional argument in function definition
    # "NPY",      # NumPy stuff
    # "FBT003",   # Boolean positional value in function call
    # "PT011",    # `pytest.raises(Exception)` is too broad
    # "B017",     # Do not assert blind exception: `Exception`
    # "PT012",    # `pytest.raises()` block should contain a single simple statement
]

[tool.ruff.lint.per-file-ignores]
# "tests/*" = ["S101"]     # Ignore assert statements in tests
# "!src/**.py" = ["D"]     # Ignore docstrings everywhere except for the `src/` directory.
"examples/*" = ["T201"]  # Ignore print statement violations in examples

[tool.ruff.lint.pydocstyle]
convention = "google"

[tool.pyright]
typeCheckingMode = "strict"
````

## File: README.md
````markdown
# gp-diagnostics

**gp-diagnostics** is a Python library for diagnosing Gaussian Process (GP) models. It provides tools for advanced
cross-validation, residual analysis, and visualization to support the evaluation and interpretation of GP regression
models.

![Example plot](examples/normal_q_q_plot_of_transformed_loo_residuals.png)

---

## Key Features

- **Cross-Validation**  
  - **Leave-One-Out (LOO)** or **Multifold** CV for GP regressions with optional fixed observational noise.  
  - Fast computation of CV residuals (including covariance) based on the approach by
  [Ginsbourger and Schaerer (2021)](https://arxiv.org/abs/2101.03108).  
  - Cholesky-based implementations with fallback checks for matrix definiteness.

- **GP Evaluation Metrics**  
  - Functions to compute **log marginal likelihood**, **pseudo-likelihood** (for CV), and **mean squared error**.  
  - Easy integration with existing GP code to measure model performance and residual normality.

- **Diagnostic Plots**
  - **Histogram of residuals**, **QQ plots**, **predictions vs. true** values, and more.  
  - Interactive **Plotly** figures for easy exploration and diagnostics.

---

## Installation

Below are two ways to install **gp-diagnostics** into your current project (using `pip` or
[`uv`](https://docs.astral.sh/uv/)). Since the package is not yet published on PyPI, you must install directly from this
repository.

---

### Option 1: Install with `uv` (recommended)

```bash
uv add "gp-diagnostics @ git+https://github.com/dnv-opensource/gp-diagnostics"
```

### Option 2: Install with `pip`

```bash
pip install git+https://github.com/dnv-opensource/gp-diagnostics
```

> **Note**
> - The two options above install **only** the core dependencies required to use the package, see below for development
>   installation.

## Development Installation

If you plan to contribute to the package, or need the development dependencies (e.g. to run the examples), you should do 
the following:

```bash
git clone https://github.com/dnv-opensource/gp-diagnostics.git
cd gp-diagnostics
uv sync
```

---

## Usage

Here is a brief example demonstrating leave-one-out (LOO) cross-validation on a Gaussian Process model.  
For a fully worked example (including multifold CV and detailed plotting), see
[examples/example.py](examples/example.py).

```python
import gpytorch
import torch

# gp-diagnostics modules
from gp_diagnostics.cv import loo
from gp_diagnostics.metrics import evaluate_GP

# 1. Build synthetic covariance matrix (K) & data (Y). For example, a simple RBF or Matern GP:
N = 10
x = torch.linspace(0, 1, N).view(-1, 1)
kernel = gpytorch.kernels.ScaleKernel(gpytorch.kernels.MaternKernel(nu=2.5))
K = kernel(x).to_dense()  # NxN prior covariance
Y = torch.distributions.MultivariateNormal(torch.zeros(N), K).sample()

# 2. (Optional) Add observational noise
noise_var = 0.01

# 3. Perform LOO cross-validation
loo_mean, loo_cov, residuals_transformed = loo(K.detach().numpy(), Y.detach().numpy(), noise_variance=noise_var)

# 4. Compute diagnostic metrics
metrics = evaluate_GP(K.detach().numpy(), Y.detach().numpy(), noise_variance=noise_var)
print("Log Marginal Likelihood:", metrics["log_marginal_likelihood"])
print("Pseudo-Likelihood (CV): ", metrics["log_pseudo_likelihood"])
print("MSE:", metrics["MSE"])
```

---

## Core Modules

**1. `cv.py`**  
- `loo()`: Leave-One-Out CV residuals and covariance.  
- `multifold()`: Multifold CV for grouped data (e.g., multiple time series).  
- Both handle GP prior covariance plus optional noise variance.

**2. `metrics.py`**  
- `evaluate_GP()`: Calculates log marginal likelihood, pseudo-likelihood, and MSE.  
- `log_prob_normal()` and `log_prob_standard_normal()`: Useful log probability helpers.

**3. `plots.py`**  
- Plotly-based functions to visualize residuals (histogram, QQ), predictive intervals, etc.  
- `qq_residuals()`, `hist_residuals()`, `pred_vs_error()`, and more for quick diagnostics.

**4. `utils`**  
- **`checks.py`**: Validations (e.g., numeric array, lower-triangular).  
- **`linalg.py`**: Cholesky utilities, triangular solves, matrix inversion.  
- **`stats.py`**: QQ data generation, partitioning data folds, etc.

---

## Contributing

Please feel free to open [issues](https://github.com/dnv-opensource/gp-diagnostics/issues) or submit pull requests if
you have ideas for improvements or bug fixes.  

For local development:

1. [Fork](https://github.com/dnv-opensource/gp-diagnostics/fork) or clone the repository.  
2. Create a feature branch: `git checkout -b feature/new-idea`.  
3. Implement your changes and add tests.  
4. Submit a pull request to `main` when ready.

---

## Testing

All tests live under the `tests/` directory. To run them:

```bash
# With pytest:
pytest

# Or with tox (runs multiple Python versions, if available):
tox
```

---

## Authors

- Christian Agrell ([christian.agrell@dnv.com](mailto\:christian.agrell@dnv.com))
- Magnus Kristiansen ([magnus.kristiansen@dnv.com](mailto\:magnus.kristiansen@dnv.com))

---

## License

[MIT License](LICENSE)  
&copy; 2024 [DNV](https://www.dnv.com). See [LICENSE](LICENSE) for details.

---
````

## File: src/gp_diagnostics/__init__.py
````python
"""gp_diagnostics package: Tools for diagnosing Gaussian Process models."""
````

## File: src/gp_diagnostics/cv.py
````python
"""Implements leave-one-out (LOO) and multifold cross-validation for Gaussian Process models.

Includes optional fixed observational noise and uses Cholesky-based formulas for fast
calculation of residuals, following Ginsbourger and Schaerer (2021).
"""

from __future__ import annotations

__all__ = [
    "check_folds_indices",
    "check_lower_triangular",
    "check_numeric_array",
    "loo",
    "loo_cholesky",
    "multifold",
    "multifold_cholesky",
]

import itertools
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from numpy.typing import NDArray

from gp_diagnostics.utils import checks
from gp_diagnostics.utils.linalg import (
    chol_inv,
    mulinv_solve,
    triang_solve,
    try_chol,
)


def multifold(
    K: NDArray[np.float64],
    Y_train: NDArray[np.float64],
    folds: list[list[int]],
    *,
    noise_variance: float = 0,
    check_args: bool = True,
) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]] | tuple[None, None, None]:
    """Compute multifold CV residuals for GP regression with optional noise variance.

    Args:
        K: GP prior covariance matrix, shape (n_samples, n_samples).
        Y_train: Training observations, shape (n_samples,).
        folds: List of list-index partitions. Each sublist is one fold's indices.
        noise_variance: Observational noise variance (0 = noiseless).
        check_args: If True, validates inputs (may raise assertions).

    Returns:
        (mean, cov, residuals_transformed), or (None, None, None) if Cholesky fails.
         - mean: Mean of multifold CV residuals, shape (n_samples,).
         - cov: Covariance matrix of multifold CV residuals, shape (n_samples, n_samples).
         - residuals_transformed: Residuals mapped into standard normal space, shape (n_samples,).

    Reference:
        D. Ginsbourger and C. Schaerer (2021). Fast calculation of Gaussian Process multiple-fold
        crossvalidation residuals and their covariances. arXiv:2101.03108
    """
    if check_args:
        check_numeric_array(Y_train, 1, "Y_train")
        check_numeric_array(K, 2, "K")
        assert K.shape[0] == Y_train.shape[0] and K.shape[1] == Y_train.shape[0], (
            f"The size of K {K.shape} is not compatible with Y_train {Y_train.shape}"
        )
        assert noise_variance >= 0, "noise_variance must be non-negative"
        check_folds_indices(folds, Y_train.shape[0])

    L = try_chol(K, noise_variance, "multifold")
    if L is None:
        return None, None, None

    return multifold_cholesky(L, Y_train, folds, check_args=False)


def multifold_cholesky(
    L: NDArray[np.float64],
    Y_train: NDArray[np.float64],
    folds: list[list[int]],
    *,
    check_args: bool = True,
) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
    """Compute multifold CV residuals given the Cholesky factor of the covariance matrix.

    Args:
        L: Lower-triangular Cholesky factor, shape (n_samples, n_samples).
        Y_train: Training observations, shape (n_samples,).
        folds: List of index partitions. Each sublist is one fold's indices.
        check_args: Whether to validate inputs.

    Returns:
        (mean, cov, residuals_transformed):
         - mean: Mean of multifold CV residuals.
         - cov: Covariance of multifold CV residuals.
         - residuals_transformed: Residuals mapped into standard normal space.

    Reference:
        D. Ginsbourger and C. Schaerer (2021). Fast calculation of Gaussian Process multiple-fold
        crossvalidation residuals and their covariances. arXiv:2101.03108
    """
    N_folds = len(folds)
    N_training = Y_train.shape[0]

    if check_args:
        check_lower_triangular(L, "L")
        check_numeric_array(Y_train, 1, "Y_train")
        check_folds_indices(folds, N_training)

    D = np.zeros((N_training, N_training), dtype=np.float64)
    D_inv_mean = np.zeros(N_training, dtype=np.float64)
    mean = np.zeros(N_training, dtype=np.float64)

    K_inv = chol_inv(L)
    K_inv_Y = mulinv_solve(L, Y_train)

    for i in range(N_folds):
        idx = np.ix_(folds[i], folds[i])
        block_chol = np.linalg.cholesky(K_inv[idx])
        D[idx] = chol_inv(block_chol)
        mean[folds[i]] = mulinv_solve(block_chol, K_inv_Y[folds[i]])
        D_inv_mean[folds[i]] = K_inv[idx].dot(mean[folds[i]])

    alpha = triang_solve(L, D)
    cov = alpha.T.dot(alpha)
    residuals_transformed = L.T.dot(D_inv_mean)
    return mean, cov, residuals_transformed


def loo(
    K: NDArray[np.float64],
    Y_train: NDArray[np.float64],
    *,
    noise_variance: float = 0,
    check_args: bool = True,
) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]] | tuple[None, None, None]:
    """Compute Leave-One-Out (LOO) CV residuals for GP regression with optional noise variance.

    Args:
        K: GP prior covariance matrix, shape (n_samples, n_samples).
        Y_train: Training observations, shape (n_samples,).
        noise_variance: Observational noise variance (0 = noiseless).
        check_args: If True, validates inputs (may raise assertions).

    Returns:
        (mean, cov, residuals_transformed) or (None, None, None) if Cholesky fails:
         - mean: Mean of LOO residuals.
         - cov: Covariance matrix of LOO residuals.
         - residuals_transformed: LOO residuals mapped into standard normal space.
    """
    if check_args:
        check_numeric_array(Y_train, 1, "Y_train")
        check_numeric_array(K, 2, "K")
        assert K.shape[0] == Y_train.shape[0] and K.shape[1] == Y_train.shape[0], (
            f"The size of K {K.shape} is not compatible with Y_train {Y_train.shape}"
        )
        assert noise_variance >= 0, "noise_variance must be non-negative"

    L = try_chol(K, noise_variance, "loo")
    if L is None:
        return None, None, None

    return loo_cholesky(L, Y_train, check_args=False)


def loo_cholesky(
    L: NDArray[np.float64],
    Y_train: NDArray[np.float64],
    *,
    check_args: bool = True,
) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
    """Compute LOO residuals given the Cholesky factor of the covariance matrix.

    Args:
        L: Lower-triangular Cholesky factor (n_samples x n_samples).
        Y_train: Training observations, shape (n_samples,).
        check_args: If True, validates arguments.

    Returns:
        (mean, cov, residuals_transformed):
         - mean: Mean of LOO residuals.
         - cov: Covariance of LOO residuals.
         - residuals_transformed: LOO residuals mapped to standard normal space.

    References:
        O. Dubrule. Cross validation of kriging in a unique neighborhood.
        Journal of the International Association for Mathematical Geology, 15 (6):687-699, 1983.
    """
    if check_args:
        check_lower_triangular(L, "L")
        check_numeric_array(Y_train, 1, "Y_train")

    K_inv = chol_inv(L)

    var = 1.0 / K_inv.diagonal()
    mean = mulinv_solve(L, Y_train) * var

    D = np.eye(var.shape[0], dtype=np.float64) * var
    alpha = triang_solve(L, D)
    cov = alpha.T.dot(alpha)

    factor = np.eye(var.shape[0], dtype=np.float64) * (1.0 / var)
    residuals_transformed = L.T.dot(factor).dot(mean)

    return mean, cov, residuals_transformed


def check_folds_indices(folds: list[list[int]], n_max: int) -> None:
    """Check that folds is a valid partition of indices [0..n_max-1].

    Args:
        folds: List of sublists of integer indices.
        n_max: Total number of samples to index.

    Raises:
        AssertionError: If folds do not form a valid partition of range(n_max).
    """
    assert isinstance(folds, list), "'folds' must be a list of lists of integers"
    assert all(isinstance(x, list) for x in folds), "'folds' must be a list of lists of integers"
    assert [] not in folds, "'folds' must not contain empty subsets"

    all_elements_set = set(itertools.chain(*folds))
    assert all(np.issubdtype(type(x), np.integer) for x in all_elements_set), (
        "'folds' must contain only integer indices"
    )
    assert all_elements_set == set(range(n_max)), "The indices in 'folds' must form a partition of range(n_max)"


def check_lower_triangular(arr: NDArray[np.float64], argname: str = "arr") -> None:
    """Check that arr is a lower-triangular 2D numeric numpy array.

    Args:
        arr: Array to check.
        argname: Name of the argument for error messages.

    Raises:
        AssertionError: If arr is not a valid lower-triangular numeric array.
    """
    assert checks.is_numeric_np_array(arr), f"{argname} must be a numpy array with numeric elements"
    assert checks.is_square(arr), f"{argname} must be square"
    assert checks.is_lower_triang(arr), f"{argname} must be lower triangular"


def check_numeric_array(arr: NDArray[np.float64], dim: int, argname: str = "arr") -> None:
    """Check that arr is a numeric numpy array of specified dimension.

    Args:
        arr: Array to check.
        dim: Required dimensionality (1 for vector, 2 for matrix, etc.).
        argname: Argument name for error messages.

    Raises:
        AssertionError: If arr is not a numeric array of the requested dimension.
    """
    assert checks.is_numeric_np_array(arr), f"{argname} must be a numeric numpy array"
    assert len(arr.shape) == dim, f"{argname} must be {dim}-dimensional"


def _multifold_inv(
    K: NDArray[np.float64],
    Y_train: NDArray[np.float64],
    folds: list[list[int]],
) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
    """Reference implementation using matrix inverse for testing multifold CV.

    Args:
        K: Covariance matrix, shape (n_samples, n_samples).
        Y_train: Observations, shape (n_samples,).
        folds: List of fold indices.

    Returns:
        (mean, cov, residuals_transformed) for multifold CV, used for testing only.
    """
    K_inv = np.linalg.inv(K)
    L = np.linalg.cholesky(K)

    N_training = Y_train.shape[0]
    N_folds = len(folds)

    K_inv_Y = K_inv.dot(Y_train)

    D = np.zeros((N_training, N_training), dtype=np.float64)
    D_inv = np.zeros((N_training, N_training), dtype=np.float64)
    mean = np.zeros(N_training, dtype=np.float64)
    D_inv_mean = np.zeros(N_training, dtype=np.float64)

    for i in range(N_folds):
        idx = np.ix_(folds[i], folds[i])
        block_inv = np.linalg.inv(K_inv[idx])
        D[idx] = block_inv
        D_inv[idx] = K_inv[idx]

        mean[folds[i]] = block_inv.dot(K_inv_Y[folds[i]])
        D_inv_mean[folds[i]] = D_inv[idx].dot(mean[folds[i]])

    cov = np.linalg.multi_dot([D, K_inv, D])
    residuals_transformed = L.T.dot(D_inv_mean)
    return mean, cov, residuals_transformed
````

## File: src/gp_diagnostics/metrics.py
````python
"""Provides functions for evaluating GP models via log marginal likelihood, pseudo-likelihood, and MSE.

Also includes convenience methods for log probability calculations under normal distributions.
"""

from __future__ import annotations

__all__ = ["evaluate_GP", "evaluate_GP_cholesky", "log_prob_normal", "log_prob_standard_normal"]

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from numpy.typing import NDArray

from gp_diagnostics.cv import (
    check_folds_indices,
    check_lower_triangular,
    check_numeric_array,
    loo_cholesky,
    multifold_cholesky,
)
from gp_diagnostics.utils.linalg import triang_solve, try_chol


def evaluate_GP(
    K: NDArray[np.float64],
    Y_train: NDArray[np.float64],
    *,
    folds: list[list[int]] | None = None,
    noise_variance: float = 0,
    check_args: bool = True,
) -> dict[str, NDArray[np.float64] | float]:
    """Computes various GP evaluation metrics for the given covariance matrix and data.

    Optionally uses multifold CV if folds are provided, otherwise does LOO.

    Args:
        K: Prior covariance matrix, shape (n_samples, n_samples).
        Y_train: Training observations, shape (n_samples,).
        folds: List of index subsets for multifold CV. If None, does LOO.
        noise_variance: Observational noise variance (>= 0).
        check_args: If True, checks input validity.

    Returns:
        A dict with:
         - "log_marginal_likelihood": float
         - "log_pseudo_likelihood": float (sum of log-prob in the standard normal space for CV residuals)
         - "MSE": float
         - "residuals_mean": array of shape (n_samples,)
         - "residuals_var": array of shape (n_samples,)
         - "residuals_transformed": array of shape (n_samples,)
    """
    if check_args:
        check_numeric_array(Y_train, 1, "Y_train")
        check_numeric_array(K, 2, "K")
        assert K.shape[0] == Y_train.shape[0] and K.shape[1] == Y_train.shape[0], (
            f"The size of K {K.shape} is not compatible with Y_train {Y_train.shape}"
        )
        assert noise_variance >= 0, "noise_variance must be non-negative"
        if folds is not None:
            check_folds_indices(folds, Y_train.shape[0])

    L = try_chol(K, noise_variance, "evaluate_GP")
    if L is None:
        return {}

    return evaluate_GP_cholesky(L, Y_train, folds=folds, check_args=False)


def evaluate_GP_cholesky(
    L: NDArray[np.float64],
    Y_train: NDArray[np.float64],
    *,
    folds: list[list[int]] | None = None,
    check_args: bool = True,
) -> dict[str, NDArray[np.float64] | float]:
    """Performs GP evaluation metrics given a Cholesky factor of (K + noise*I).

    Args:
        L: Lower-triangular factor, shape (n_samples, n_samples).
        Y_train: Observations, shape (n_samples,).
        folds: If not None, multifold CV indices; else uses LOO.
        check_args: Whether to verify argument shapes.

    Returns:
        A dictionary of metrics:
         - "log_marginal_likelihood"
         - "log_pseudo_likelihood"
         - "MSE"
         - "residuals_mean"
         - "residuals_var"
         - "residuals_transformed"
    """
    if check_args:
        check_lower_triangular(L, "L")
        check_numeric_array(Y_train, 1, "Y_train")
        if folds is not None:
            check_folds_indices(folds, Y_train.shape[0])

    res: dict[str, NDArray[np.float64] | float] = {}
    if folds is not None:
        mean, cov, residuals_transformed = multifold_cholesky(L, Y_train, folds, check_args=False)
    else:
        mean, cov, residuals_transformed = loo_cholesky(L, Y_train, check_args=False)

    res["log_marginal_likelihood"] = log_prob_normal(L, Y_train)
    res["log_pseudo_likelihood"] = log_prob_standard_normal(residuals_transformed)
    res["MSE"] = float(np.linalg.norm(mean))

    res["residuals_mean"] = mean
    res["residuals_var"] = cov.diagonal()
    res["residuals_transformed"] = residuals_transformed

    return res


def log_prob_normal(L: NDArray[np.float64], Y: NDArray[np.float64]) -> float:
    """Computes log probability of data Y under Gaussian with covariance L*L^T.

    Args:
        L: Lower-triangular factor, shape (n_samples, n_samples).
        Y: Data vector, shape (n_samples,).

    Returns:
        Scalar log probability of Y.
    """
    a = triang_solve(L, Y)
    return -0.5 * np.linalg.norm(a) ** 2 - np.log(L.diagonal()).sum() - (Y.shape[0] / 2) * np.log(2 * np.pi)


def log_prob_standard_normal(Y: NDArray[np.float64]) -> float:
    """Computes log probability of data Y under a standard normal distribution.

    Args:
        Y: Data vector, shape (n_samples,).

    Returns:
        The scalar log probability of Y under N(0, I).
    """
    return -0.5 * np.linalg.norm(Y) ** 2 - (Y.shape[0] / 2) * np.log(2 * np.pi)
````

## File: src/gp_diagnostics/plots.py
````python
"""Plotly-based functions for visualizing residuals, predictions, QQ plots, histograms, etc.

Used to aid in GP diagnostics.
"""

from __future__ import annotations

__all__ = [
    "error_scatter",
    "gp_diagnostics",
    "hist_residuals",
    "pred_vs_error",
    "pred_vs_error_perc",
    "qq_residuals",
]

from typing import TYPE_CHECKING

import numpy as np
import plotly
import plotly.graph_objs as go
from plotly.subplots import make_subplots
from scipy.stats import gaussian_kde, norm

if TYPE_CHECKING:
    import pandas as pd
    from numpy.typing import NDArray

from gp_diagnostics.utils.stats import snorm_qq


def hist_residuals(
    y_pred_mean: NDArray[np.float64],
    y_pred_var: NDArray[np.float64],
    y_test: NDArray[np.float64],
    *,
    title: str = "",
    showlegend: bool = True,
) -> plotly.graph_objs.Figure:
    """Creates a Plotly histogram of standardized residuals with normal & KDE overlays.

    Args:
        y_pred_mean: Predicted means, shape (n_samples,).
        y_pred_var: Predicted variances, shape (n_samples,).
        y_test: True target values, shape (n_samples,).
        title: Plot title.
        showlegend: Whether to show the legend.

    Returns:
        A Plotly figure with a histogram of residuals.
    """
    clr = "rgb(105, 144, 193)"
    y_pred_std = np.sqrt(y_pred_var)
    residuals_y = (y_pred_mean - y_test) / y_pred_std

    hist = go.Histogram(
        x=residuals_y,
        histnorm="probability density",
        marker_color=clr,
        opacity=0.5,
        name="residuals",
    )

    x_min, x_max = float(residuals_y.min()), float(residuals_y.max())
    x_points = np.linspace(x_min, x_max, 100)

    normal_dens = go.Scatter(
        x=x_points,
        y=norm.pdf(x_points),
        mode="lines",
        name="standard normal density",
        line={"color": "black", "width": 1},
    )
    kde = gaussian_kde(residuals_y)
    kde_dens = go.Scatter(
        x=x_points,
        y=kde(x_points),
        mode="lines",
        name="residuals kde",
        line={"color": clr, "width": 1},
    )

    layout = go.Layout(
        title=title,
        showlegend=showlegend,
        autosize=False,
        width=700,
        height=600,
        xaxis={"title": "Standardised error"},
        yaxis={"title": "Density"},
        legend={"yanchor": "top", "y": 0.99, "xanchor": "left", "x": 0.01},
    )

    data = [hist, kde_dens, normal_dens]
    return go.Figure(data=data, layout=layout)


def qq_residuals(
    y_pred_mean: NDArray[np.float64],
    y_pred_var: NDArray[np.float64],
    y_test: NDArray[np.float64],
    *,
    title: str = "",
    showlegend: bool = True,
) -> plotly.graph_objs.Figure:
    """Creates a Plotly QQ-plot of standardized residuals vs. standard normal quantiles.

    Args:
        y_pred_mean: Predicted means, shape (n_samples,).
        y_pred_var: Predicted variances, shape (n_samples,).
        y_test: True target values, shape (n_samples,).
        title: Plot title.
        showlegend: Whether to show the legend.

    Returns:
        A Plotly Figure containing the QQ-plot.
    """
    y_pred_std = np.sqrt(y_pred_var)
    residuals_y = (y_pred_mean - y_test) / y_pred_std

    q_sample, q_snorm, q_snorm_upper, q_snorm_lower = snorm_qq(residuals_y)

    qq_scatter = go.Scatter(
        x=q_snorm,
        y=q_sample,
        mode="markers",
        marker={"size": 6, "color": "rgb(105, 144, 193)"},
        name="Data",
    )
    qq_upper = go.Scatter(
        x=q_snorm_upper,
        y=q_sample,
        mode="lines",
        line={"color": "rgb(150, 150, 150)", "dash": "dot"},
        name="95% confidence band",
        legendgroup="conf",
    )
    qq_lower = go.Scatter(
        x=q_snorm_lower,
        y=q_sample,
        mode="lines",
        line={"color": "rgb(150, 150, 150)", "dash": "dot"},
        name="lower",
        legendgroup="conf",
        showlegend=False,
    )

    minval = float(min(q_snorm.min(), q_sample.min()))
    maxval = float(max(q_snorm.max(), q_sample.max()))

    line = go.Scatter(
        x=[minval, maxval],
        y=[minval, maxval],
        mode="lines",
        line={"color": "rgb(0, 0, 0)", "dash": "dash"},
        name="x = y",
    )

    layout = go.Layout(
        title=title,
        showlegend=showlegend,
        autosize=False,
        width=700,
        height=600,
        xaxis={
            "title": "Standard normal quantiles",
            "range": [q_snorm.min() - 0.2, q_snorm.max() + 0.2],
        },
        yaxis={"title": "Sample quantiles"},
    )

    data = [qq_scatter, qq_upper, qq_lower, line]
    return go.Figure(data=data, layout=layout)


def pred_vs_error(
    y_pred_mean: NDArray[np.float64],
    y_pred_var: NDArray[np.float64],
    y_test: NDArray[np.float64],
    *,
    title: str = "",
    showlegend: bool = True,
) -> plotly.graph_objs.Figure:
    """Creates a scatter plot comparing predicted vs. true values, with 95% error bars.

    Args:
        y_pred_mean: Predicted means, shape (n_samples,).
        y_pred_var: Predicted variances, shape (n_samples,).
        y_test: True target values, shape (n_samples,).
        title: Plot title.
        showlegend: Whether to show the legend.

    Returns:
        A Plotly Figure showing predicted vs. actual + 95% intervals.
    """
    minval = float(min(y_pred_mean.min(), y_test.min()))
    maxval = float(max(y_pred_mean.max(), y_test.max()))

    y_pred_std = np.sqrt(y_pred_var)
    num_std = 1.959963984540054

    pred_scatter = go.Scatter(
        x=y_test,
        y=y_pred_mean,
        mode="markers",
        marker={"size": 6, "color": "rgb(105, 144, 193)"},
        name="Prediction",
    )

    pred_bars = go.Scatter(
        x=y_test,
        y=y_pred_mean,
        mode="markers",
        marker={"size": 6, "color": "rgb(105, 144, 193)"},
        error_y={
            "type": "data",
            "array": y_pred_std * num_std,
            "visible": True,
            "color": "rgb(105, 144, 193)",
        },
        name="95% intervals",
    )

    line = go.Scatter(
        x=[minval, maxval],
        y=[minval, maxval],
        mode="lines",
        line={"color": "rgb(0, 0, 0)", "dash": "dash"},
        name="x = y",
    )

    layout = go.Layout(
        title=title,
        showlegend=showlegend,
        autosize=False,
        width=700,
        height=600,
        xaxis={"title": "True value"},
        yaxis={"title": "Predicted value"},
    )

    data = [pred_bars, pred_scatter, line]
    return go.Figure(data=data, layout=layout)


def pred_vs_error_perc(
    y_pred_mean: NDArray[np.float64],
    y_pred_perc_lower: NDArray[np.float64],
    y_pred_perc_upper: NDArray[np.float64],
    y_test: NDArray[np.float64],
    conf_interval: float,
    *,
    title: str = "",
    showlegend: bool = True,
) -> plotly.graph_objs.Figure:
    """Creates a scatter plot of predicted vs. true values with percentile error bars.

    Args:
        y_pred_mean: Predicted means, shape (n_samples,).
        y_pred_perc_lower: Lower percentile predictions, shape (n_samples,).
        y_pred_perc_upper: Upper percentile predictions, shape (n_samples,).
        y_test: True target values, shape (n_samples,).
        conf_interval: The percentile coverage (e.g. 95).
        title: Plot title.
        showlegend: Whether to show the legend.

    Returns:
        A Plotly Figure with predicted vs. actual scatter + percentile error bars.
    """
    minval = float(min(y_pred_mean.min(), y_test.min()))
    maxval = float(max(y_pred_mean.max(), y_test.max()))

    pred_scatter = go.Scatter(
        x=y_test,
        y=y_pred_mean,
        mode="markers",
        marker={"size": 6, "color": "rgb(105, 144, 193)"},
        name="Prediction",
    )

    pred_bars = go.Scatter(
        x=y_test,
        y=(y_pred_perc_upper + y_pred_perc_lower) / 2.0,
        mode="markers",
        marker={"size": 0, "color": "rgb(105, 144, 193)", "opacity": 0},
        error_y={
            "type": "data",
            "array": (y_pred_perc_upper - y_pred_perc_lower) / 2.0,
            "visible": True,
            "color": "rgb(105, 144, 193)",
        },
        name=f"{conf_interval}% intervals",
    )

    line = go.Scatter(
        x=[minval, maxval],
        y=[minval, maxval],
        mode="lines",
        line={"color": "rgb(0, 0, 0)", "dash": "dash"},
        name="x = y",
    )

    layout = go.Layout(
        title=title,
        showlegend=showlegend,
        autosize=False,
        width=700,
        height=600,
        xaxis={"title": "True value"},
        yaxis={"title": "Predicted value"},
    )

    data = [pred_bars, pred_scatter, line]
    return go.Figure(data=data, layout=layout)


def error_scatter(
    x: NDArray[np.float64],
    errors: NDArray[np.float64],
    *,
    title: str = "",
    x_label: str = "x",
    y_label: str = "Standardized errors",
    showlegend: bool = True,
) -> plotly.graph_objs.Figure:
    """Creates a scatter plot of standardized errors vs. some variable x, with 95% reference lines.

    Args:
        x: Horizontal-axis variable, shape (n_samples,).
        errors: Standardized error values, shape (n_samples,).
        title: Plot title.
        x_label: Label for x-axis.
        y_label: Label for y-axis.
        showlegend: Whether to display the legend.

    Returns:
        A Plotly Figure of the scatter + reference lines at ±1.96 std dev.
    """
    scatter_points = go.Scatter(
        x=x,
        y=errors,
        mode="markers",
        marker={"size": 6, "color": "rgb(105, 144, 193)"},
        name=y_label,
    )

    num_std = 1.959963984540054
    min_x = float(x.min())
    max_x = float(x.max())
    margin = (max_x - min_x) * 0.025
    x_line = [min_x - margin, max_x + margin]

    line_mid = go.Scatter(
        x=x_line,
        y=[0, 0],
        mode="lines",
        line={"color": "rgb(0, 0, 0)", "dash": "dash"},
        name="",
        showlegend=False,
    )
    line_upper = go.Scatter(
        x=x_line,
        y=[num_std, num_std],
        mode="lines",
        line={"color": "rgb(150, 150, 150)", "dash": "dot"},
        name="95% interval",
        legendgroup="conf",
    )
    line_lower = go.Scatter(
        x=x_line,
        y=[-num_std, -num_std],
        mode="lines",
        line={"color": "rgb(150, 150, 150)", "dash": "dot"},
        name="95% interval",
        legendgroup="conf",
        showlegend=False,
    )

    layout = go.Layout(
        title=title,
        showlegend=showlegend,
        autosize=False,
        width=700,
        height=600,
        xaxis={"title": x_label, "range": [x_line[0], x_line[1]]},
        yaxis={"title": y_label},
    )

    data = [scatter_points, line_mid, line_upper, line_lower]
    return go.Figure(data=data, layout=layout)


def gp_diagnostics(
    data: pd.DataFrame,
    y_name: str,
    *,
    plot_labels: dict[str, str] | None = None,
    subplots: bool = True,
) -> list[plotly.graph_objs.Figure]:
    """Creates a set of Plotly figures for GP diagnostics (QQ, pred vs. true, error scatter).

    Args:
        data: Pandas DataFrame with columns for inputs and outputs. Must contain {y_name}_true, {y_name}_mean, and
              {y_name}_var.
        y_name: Base name of the output variable.
        plot_labels: Optional dict to rename variables in plot titles.
        subplots: If True, returns combined subplots for some figures.

    Returns:
        A list of Plotly figures for GP diagnostics, or subplots if requested.

    Raises:
        ValueError: If required columns for y_name are missing in data.
    """
    if plot_labels is None:
        plot_labels = {}

    outputnames = [f"{y_name}_true", f"{y_name}_mean", f"{y_name}_var"]
    inputnames = [col for col in data.columns if col not in outputnames]

    missing_cols = [col for col in outputnames if col not in data.columns]
    if missing_cols:
        msg = f"DataFrame is missing required columns for '{y_name}'.\nMissing columns: {missing_cols}"
        raise ValueError(msg)

    y_pred_mean = data[f"{y_name}_mean"].to_numpy(dtype=np.float64)
    y_pred_var = data[f"{y_name}_var"].to_numpy(dtype=np.float64)
    y_test = data[f"{y_name}_true"].to_numpy(dtype=np.float64)
    y_pred_std = np.sqrt(y_pred_var)
    residuals_y = (y_pred_mean - y_test) / y_pred_std

    dict_varnames = dict(plot_labels)
    for name in data.columns:
        if name not in dict_varnames:
            dict_varnames[name] = name
    if y_name not in dict_varnames:
        dict_varnames[y_name] = y_name

    # QQ plot
    fig_qq = qq_residuals(y_pred_mean, y_pred_var, y_test, title="Normal QQ of errors", showlegend=False)

    # Prediction vs test
    fig_pred_vs_err = pred_vs_error(y_pred_mean, y_pred_var, y_test, title="Prediction vs test", showlegend=False)

    # Standardised error scatter
    figs_errorscatter = []
    x_vals = y_pred_mean
    x_label = f"GP mean E[{dict_varnames[y_name]}]"
    figs_errorscatter.append(
        error_scatter(
            x_vals,
            residuals_y,
            title="Standardised errors vs mean",
            x_label=x_label,
            y_label="Standardised errors",
            showlegend=False,
        )
    )

    x_vals = y_pred_var
    x_label = f"GP variance Var[{dict_varnames[y_name]}]"
    figs_errorscatter.append(
        error_scatter(
            x_vals,
            residuals_y,
            title="Standardised errors vs variance",
            x_label=x_label,
            y_label="Standardised errors",
            showlegend=False,
        )
    )

    for name in inputnames:
        x_vals = data[name].to_numpy(dtype=np.float64)
        x_label = dict_varnames[name]
        figs_errorscatter.append(
            error_scatter(
                x_vals,
                residuals_y,
                title=f"Standardised errors vs {x_label}",
                x_label=x_label,
                y_label="Standardised errors",
                showlegend=False,
            )
        )

    if not subplots:
        return [fig_qq, fig_pred_vs_err, *figs_errorscatter]

    fig1 = make_subplots(rows=1, cols=2, subplot_titles=("Prediction vs test", "Standardised errors QQ"))
    for trace in fig_pred_vs_err.data:
        fig1.add_trace(trace, row=1, col=1)
    for trace in fig_qq.data:
        fig1.add_trace(trace, row=1, col=2)

    fig1.update_xaxes(fig_pred_vs_err.layout.xaxis, row=1, col=1)
    fig1.update_yaxes(fig_pred_vs_err.layout.yaxis, row=1, col=1)
    fig1.update_xaxes(fig_qq.layout.xaxis, row=1, col=2)
    fig1.update_yaxes(fig_qq.layout.yaxis, row=1, col=2)
    fig1.update_layout(showlegend=False)

    numcols = 3
    numplots = len(figs_errorscatter)
    numrows = (numplots + numcols - 1) // numcols

    fig3 = make_subplots(rows=numrows, cols=numcols)
    idx = 0
    for i in range(numrows):
        for j in range(numcols):
            if idx < numplots:
                for trc in figs_errorscatter[idx].data:
                    fig3.add_trace(trc, row=i + 1, col=j + 1)
                fig3.update_xaxes(
                    title=figs_errorscatter[idx].layout.xaxis.title,
                    row=i + 1,
                    col=j + 1,
                )
            idx += 1

    fig3.update_layout(showlegend=False, title="Standardised errors", height=330 * numrows, width=950)

    return [fig1, fig3]
````

## File: src/gp_diagnostics/utils/__init__.py
````python
"""gp_diagnostics.utils: Utility subpackage for common helper functions and classes."""
````

## File: src/gp_diagnostics/utils/checks.py
````python
"""Utilities for checking array properties such as 'numeric', 'square', and 'lower triangular'."""

from __future__ import annotations

__all__ = ["is_lower_triang", "is_numeric_np_array", "is_square"]

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from numpy.typing import NDArray


def is_numeric_np_array(arr: object) -> bool:
    """Check if arr is a numpy array with numeric dtype.

    Args:
        arr: Object to verify.

    Returns:
        True if arr is a numpy array with numeric type, else False.
    """
    if not isinstance(arr, np.ndarray):
        return False
    return arr.dtype.kind in {"b", "u", "i", "f", "c"}


def is_square(arr: NDArray[np.float64]) -> bool:
    """Check if arr is a 2D square numpy array.

    Args:
        arr: A numpy array to check.

    Returns:
        True if arr is 2D and arr.shape[0] == arr.shape[1], else False.
    """
    if arr.ndim != 2:  # noqa: PLR2004
        return False
    return arr.shape[0] == arr.shape[1]


def is_lower_triang(arr: NDArray[np.float64]) -> bool:
    """Check if arr is lower triangular.

    Args:
        arr: A 2D square numpy array.

    Returns:
        True if arr is lower triangular, else False.
    """
    idx = np.triu_indices_from(arr, k=1)
    return np.all(arr[idx] == 0)
````

## File: src/gp_diagnostics/utils/linalg.py
````python
"""Provides linear algebra utilities for Cholesky inversion, triangular solves, and related operations."""

from __future__ import annotations

__all__ = [
    "chol_inv",
    "mulinv_solve",
    "mulinv_solve_rev",
    "symmetrify",
    "traceprod",
    "triang_solve",
    "try_chol",
]

import warnings
from typing import TYPE_CHECKING

import numpy as np
from scipy import linalg

if TYPE_CHECKING:
    from numpy.typing import NDArray


def triang_solve(
    A: NDArray[np.float64],
    B: NDArray[np.float64],
    *,
    lower: bool = True,
    trans: bool = False,
) -> NDArray[np.float64]:
    """Solve a triangular system A*X = B or A^T*X = B, where A is triangular.

    Args:
        A: Triangular matrix of shape (N, N).
        B: Right-hand side, shape (N, NRHS).
        lower: If True, A is lower triangular; else upper.
        trans: If True, solve A^T * X = B; else A * X = B.

    Returns:
        A solution matrix X, shape (N, NRHS).
    """
    unitdiag = False
    lower_num = 1 if lower else 0
    trans_num = 1 if trans else 0
    unitdiag_num = 1 if unitdiag else 0

    A_fortran = np.asfortranarray(A)
    return linalg.lapack.dtrtrs(
        A_fortran,
        B,
        lower=lower_num,
        trans=trans_num,
        unitdiag=unitdiag_num,
    )[0]


def mulinv_solve(
    F: NDArray[np.float64],
    B: NDArray[np.float64],
    *,
    lower: bool = True,
) -> NDArray[np.float64]:
    """Solve A*X = B for A = F * F^T. Typically used when F is lower triangular.

    Args:
        F: Lower-triangular factor, shape (N, N).
        B: Right-hand side, shape (N, NRHS).
        lower: If True, F is lower triangular; else upper.

    Returns:
        The solution X, shape (N, NRHS).
    """
    tmp = triang_solve(F, B, lower=lower, trans=False)
    return triang_solve(F, tmp, lower=lower, trans=True)


def mulinv_solve_rev(
    F: NDArray[np.float64],
    B: NDArray[np.float64],
    *,
    lower: bool = True,
) -> NDArray[np.float64]:
    """Solve X*A = B for A = F*F^T. Typically used when F is lower triangular.

    Args:
        F: Lower-triangular factor, shape (N, N).
        B: Right-hand side, shape (NRHS, N).
        lower: If True, F is lower triangular; else upper.

    Returns:
        The solution X, shape (NRHS, N).
    """
    return mulinv_solve(F, B.T, lower=lower).T


def symmetrify(A: NDArray[np.float64], *, upper: bool = False) -> None:
    """Make matrix A symmetric by copying one triangle to the other in-place.

    Args:
        A: A square 2D numpy array.
        upper: If True, copy the upper triangle to the lower; else vice versa.
    """
    triu = np.triu_indices_from(A, k=1)
    if upper:
        A.T[triu] = A[triu]
    else:
        A[triu] = A.T[triu]


def chol_inv(L: NDArray[np.float64]) -> NDArray[np.float64]:
    """Return inverse of A = L * L^T, where L is lower triangular (via LAPACK dpotri).

    Args:
        L: Lower-triangular factor, shape (N, N).

    Returns:
        A_inv: The inverse of A, shape (N, N).
    """
    A_inv, _info = linalg.lapack.dpotri(L, lower=1)
    symmetrify(A_inv)
    return A_inv


def traceprod(A: NDArray[np.float64], B: NDArray[np.float64]) -> float:
    """Compute trace(A @ B) using Einstein summation.

    Args:
        A: 2D numpy array.
        B: 2D numpy array with shape (columns_A, something).

    Returns:
        Scalar trace of A*B.
    """
    return float(np.einsum("ij,ji->", A, B))


def try_chol(
    K: NDArray[np.float64],
    noise_variance: float,
    fun_name: str,
) -> NDArray[np.float64] | None:
    """Attempt Cholesky of (K + noise_variance*I). If fail, warn and return None.

    Args:
        K: Covariance matrix, shape (N, N).
        noise_variance: Noise variance to add to diagonal.
        fun_name: Label for logging/warnings.

    Returns:
        The lower-triangular Cholesky factor or None if decomposition fails.
    """
    A = K + np.eye(K.shape[0], dtype=np.float64) * noise_variance
    try:
        return np.linalg.cholesky(A)
    except np.linalg.LinAlgError:
        warnings.warn(
            f"Could not compute Cholesky in '{fun_name}'; matrix likely not PD. Returning None.",
            stacklevel=2,
        )
        try:
            min_eig = np.linalg.eig(A)[0].min()
            warnings.warn(f"Smallest eigenvalue: {min_eig}", stacklevel=2)
        except np.linalg.LinAlgError:
            warnings.warn("Could not compute smallest eigenvalue.", stacklevel=2)
        return None
````

## File: src/gp_diagnostics/utils/stats.py
````python
"""Statistical utilities for generating normal QQ data, partitioning folds, etc."""

from __future__ import annotations

__all__ = ["snorm_qq", "split_test_train_fold"]

import itertools
from typing import TYPE_CHECKING

import numpy as np
from scipy.stats import norm

if TYPE_CHECKING:
    from numpy.typing import NDArray


def snorm_qq(
    x: NDArray[np.float64],
) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
    """Calculate standard normal QQ plot data, with approximate 95% confidence bands.

    Args:
        x: 1D data array for which to compute QQ plot info.

    Returns:
        (q_sample, q_snorm, q_snorm_upper, q_snorm_lower):
         - q_sample: Sorted sample quantiles from x.
         - q_snorm: Theoretical standard normal quantiles.
         - q_snorm_upper: Upper confidence band of theoretical quantiles.
         - q_snorm_lower: Lower confidence band of theoretical quantiles.
    """
    n = len(x)
    q_sample = np.sort(x)

    p = (np.arange(n) + 0.5) / n
    q_snorm = norm.ppf(p)

    # Approx confidence band
    k = 0.895 / (np.sqrt(n) * (1 - 0.01 / np.sqrt(n) + 0.85 / n))
    q_snorm_upper = norm.ppf(p + k)
    q_snorm_lower = norm.ppf(p - k)

    return q_sample, q_snorm, q_snorm_upper, q_snorm_lower


def split_test_train_fold(
    folds: list[list[int]],
    X: NDArray[np.float64],
    i: int,
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Split X into test and train arrays based on the i-th fold indices.

    Args:
        folds: List of index subsets.
        X: Array or matrix of shape (n_samples, ...).
        i: Fold index to designate as test.

    Returns:
        (X_test, X_train) with shapes based on folds[i].
    """
    idx_test = folds[i]
    idx_train = list(itertools.chain(*[folds[j] for j in range(len(folds)) if j != i]))
    return X[idx_test], X[idx_train]
````

## File: tests/test_cv.py
````python
"""Unit tests for cv.py, focusing on correctness of LOO and multifold cross-validation residuals."""

import functools
import operator
import random

import gpytorch
import numpy as np
import pytest
import torch
from numpy.typing import NDArray

from gp_diagnostics.cv import check_folds_indices, check_lower_triangular, check_numeric_array, loo, multifold
from gp_diagnostics.utils.stats import split_test_train_fold
from tests.utils import (
    ExactGPModel,
    gpytorch_kernel_Matern,
    gpytorch_likelihood_gaussian,
    gpytorch_mean_constant,
)


def test_check_folds_indices_correct() -> None:
    """Test that check_folds_indices throws assertion when needed.

    - No assertion when input is correct
    """
    # Correct (should not do anything)
    check_folds_indices([[1, 3], [5, 6, 7], [0, 2], [4]], 8)


def test_check_folds_indices_nmax() -> None:
    """Test that check_folds_indices throws assertion when needed.

    - Wrong n_max
    """
    with pytest.raises(Exception):
        check_folds_indices([[1, 3], [5, 6, 7], [0, 2], [4]], 6)  # Wrong n_max


def test_check_folds_indices_list() -> None:
    """Test that check_folds_indices throws assertion when needed.

    - Not a list of lists
    """
    with pytest.raises(Exception):
        # Not a list of lists
        check_folds_indices([1, 3, 5, 6, 7, 0, 2, 4], 8)  # type: ignore


def test_check_folds_indices_int() -> None:
    """Test that check_folds_indices throws assertion when needed.

    - Not a list of lists of ints
    """
    with pytest.raises(Exception):
        # Wrong type: str (instead of int)
        check_folds_indices([["22"]], 8)  # type: ignore
        # Wrong type: float (instead of int)
        check_folds_indices([[1, 0.5]], 8)  # type: ignore


def test_check_folds_indices_exh() -> None:
    """Test that check_folds_indices throws assertion when needed.

    - Not exhaustive indices
    """
    with pytest.raises(Exception):
        check_folds_indices([[1, 3], [5, 7], [0, 2], [4]], 8)  # Not exhaustive indices
        check_folds_indices([[1, 3], [5, 6, 7, 9], [0, 2], [4]], 8)  # Not exhaustive indices


def test_check_folds_indices_empty() -> None:
    """Test that check_folds_indices throws assertion when needed.

    - Empty lists
    """
    with pytest.raises(Exception):
        check_folds_indices([[1, 3], [5, 6, 7], [0, 2], [], [4]], 8)


def test_check_lower_triangular() -> None:
    """Test that check_lower_triangular throws assertion when needed.

    including multiple cases in same test as check_lower_triangular already uses
    tested functions
    """
    # These should be ok
    arr = np.array([[0.2, 0, 0], [3, 2.2, 0], [1, 2, 4]])

    check_lower_triangular(arr)

    arr = np.array([[1, 0], [2, 2.2]])
    check_lower_triangular(arr)

    arr = np.array([[1]])
    check_lower_triangular(arr)

    # These should raise error
    with pytest.raises(Exception):
        check_lower_triangular(np.array([[1, 2, 2.3], [0, 2.2, 3], [0.1, 0, 4]]))

    with pytest.raises(Exception):
        check_lower_triangular(np.ones(shape=(13, 14)))

    with pytest.raises(Exception):
        check_lower_triangular(np.array([[1, 2, 2.3], [0, 0.001, "a"]]))

    with pytest.raises(Exception):
        check_lower_triangular("a")


def test_check_numeric_array() -> None:
    """Test that check_numeric_array throws assertion when needed.

    including multiple cases in same test as check_numeric_array already uses
    tested functions
    """
    # These should be ok
    check_numeric_array(np.ones(4), 1)
    check_numeric_array(np.ones(shape=(2, 4)), 2)
    check_numeric_array(np.ones(shape=(2, 4, 6)), 3)
    check_numeric_array(np.array(4), 0)

    # These should raise error
    with pytest.raises(Exception):
        check_numeric_array(np.array(4), 1)

    with pytest.raises(Exception):
        check_numeric_array(np.ones(shape=(2, 4)), 1)

    with pytest.raises(Exception):
        check_numeric_array("a", 1)

    with pytest.raises(Exception):
        check_numeric_array(np.array([1, "1"]), 1)


def test_loo_1d() -> None:
    """Test the 1d example from.

    [D. Ginsbourger and C. Schaerer (2021). Fast calculation of Gaussian Process
        multiple-fold crossvalidation residuals and their covariances.
        arXiv:2101.03108]
    """
    # Covariance matrix and observations
    Y_train = np.array(
        [
            -0.6182,
            -0.3888,
            -0.3287,
            -0.2629,
            0.3614,
            0.1442,
            -0.0374,
            -0.0546,
            -0.0056,
            0.0529,
        ]
    )

    K = np.array(
        [
            [
                9.1475710e-02,
                5.4994639e-02,
                1.8560780e-02,
                4.8545646e-03,
                1.1045272e-03,
                2.3026903e-04,
                4.5216686e-05,
                8.5007923e-06,
                1.5461808e-06,
                2.7402149e-07,
            ],
            [
                5.4994639e-02,
                9.1475710e-02,
                5.4994628e-02,
                1.8560771e-02,
                4.8545646e-03,
                1.1045267e-03,
                2.3026903e-04,
                4.5216686e-05,
                8.5007923e-06,
                1.5461794e-06,
            ],
            [
                1.8560775e-02,
                5.4994617e-02,
                9.1475710e-02,
                5.4994628e-02,
                1.8560771e-02,
                4.8545622e-03,
                1.1045267e-03,
                2.3026903e-04,
                4.5216686e-05,
                8.5007923e-06,
            ],
            [
                4.8545646e-03,
                1.8560771e-02,
                5.4994628e-02,
                9.1475710e-02,
                5.4994628e-02,
                1.8560771e-02,
                4.8545646e-03,
                1.1045275e-03,
                2.3026903e-04,
                4.5216719e-05,
            ],
            [
                1.1045275e-03,
                4.8545646e-03,
                1.8560771e-02,
                5.4994628e-02,
                9.1475710e-02,
                5.4994617e-02,
                1.8560771e-02,
                4.8545646e-03,
                1.1045275e-03,
                2.3026903e-04,
            ],
            [
                2.3026903e-04,
                1.1045267e-03,
                4.8545622e-03,
                1.8560771e-02,
                5.4994617e-02,
                9.1475710e-02,
                5.4994635e-02,
                1.8560780e-02,
                4.8545660e-03,
                1.1045275e-03,
            ],
            [
                4.5216686e-05,
                2.3026884e-04,
                1.1045267e-03,
                4.8545646e-03,
                1.8560771e-02,
                5.4994635e-02,
                9.1475710e-02,
                5.4994639e-02,
                1.8560780e-02,
                4.8545660e-03,
            ],
            [
                8.5007923e-06,
                4.5216686e-05,
                2.3026903e-04,
                1.1045275e-03,
                4.8545646e-03,
                1.8560780e-02,
                5.4994639e-02,
                9.1475710e-02,
                5.4994617e-02,
                1.8560775e-02,
            ],
            [
                1.5461794e-06,
                8.5007923e-06,
                4.5216686e-05,
                2.3026903e-04,
                1.1045272e-03,
                4.8545660e-03,
                1.8560780e-02,
                5.4994628e-02,
                9.1475710e-02,
                5.4994639e-02,
            ],
            [
                2.7402149e-07,
                1.5461808e-06,
                8.5007923e-06,
                4.5216686e-05,
                2.3026903e-04,
                1.1045275e-03,
                4.8545660e-03,
                1.8560780e-02,
                5.4994639e-02,
                9.1475710e-02,
            ],
        ]
    )

    # From paper
    LOO_residuals_transformed_true = np.array(
        [
            -2.04393396,
            -0.07086865,
            -0.81325009,
            -0.33726709,
            2.07426555,
            -0.82414653,
            -0.05894939,
            -0.10534249,
            0.10176395,
            0.18906068,
        ]
    )

    LOO_mean_true = np.array(
        [
            -0.38365906,
            0.02787939,
            0.02736787,
            -0.29997396,
            0.36816096,
            -0.11112669,
            0.0047464,
            -0.01693309,
            -0.00649325,
            0.04409083,
        ]
    )

    LOO_cov_true = np.array(
        [
            [
                5.43868914e-02,
                -2.63221730e-02,
                1.02399085e-02,
                -3.40788392e-03,
                1.09471090e-03,
                -3.50791292e-04,
                1.12430032e-04,
                -3.61632192e-05,
                1.20125414e-05,
                -4.93005518e-06,
            ],
            [
                -2.63221748e-02,
                3.40218581e-02,
                -2.04288363e-02,
                8.01187754e-03,
                -2.66015902e-03,
                8.54554935e-04,
                -2.73940881e-04,
                8.81146188e-05,
                -2.92695640e-05,
                1.20124914e-05,
            ],
            [
                1.02399066e-02,
                -2.04288289e-02,
                3.19701731e-02,
                -1.97094288e-02,
                7.72963883e-03,
                -2.56570964e-03,
                8.24513379e-04,
                -2.65259878e-04,
                8.81142623e-05,
                -3.61629245e-05,
            ],
            [
                -3.40788229e-03,
                8.01187381e-03,
                -1.97094269e-02,
                3.17552164e-02,
                -1.96319763e-02,
                7.69940997e-03,
                -2.55653122e-03,
                8.24513205e-04,
                -2.73939862e-04,
                1.12429247e-04,
            ],
            [
                1.09471008e-03,
                -2.66015716e-03,
                7.72963790e-03,
                -1.96319763e-02,
                3.17334048e-02,
                -1.96248218e-02,
                7.69941136e-03,
                -2.56570987e-03,
                8.54552840e-04,
                -3.50789691e-04,
            ],
            [
                -3.50790826e-04,
                8.54554237e-04,
                -2.56570918e-03,
                7.69940997e-03,
                -1.96248218e-02,
                3.17334011e-02,
                -1.96319763e-02,
                7.72963837e-03,
                -2.66015413e-03,
                1.09470787e-03,
            ],
            [
                1.12429749e-04,
                -2.73940503e-04,
                8.24513205e-04,
                -2.55653122e-03,
                7.69941136e-03,
                -1.96319763e-02,
                3.17552052e-02,
                -1.97094250e-02,
                8.01186915e-03,
                -3.40787880e-03,
            ],
            [
                -3.61630882e-05,
                8.81144733e-05,
                -2.65259820e-04,
                8.24513321e-04,
                -2.56571034e-03,
                7.72963930e-03,
                -1.97094250e-02,
                3.19701731e-02,
                -2.04288270e-02,
                1.02399047e-02,
            ],
            [
                1.20125114e-05,
                -2.92695440e-05,
                8.81143787e-05,
                -2.73940270e-04,
                8.54553713e-04,
                -2.66015623e-03,
                8.01187288e-03,
                -2.04288345e-02,
                3.40218581e-02,
                -2.63221730e-02,
            ],
            [
                -4.93004836e-06,
                1.20124987e-05,
                -3.61630118e-05,
                1.12429487e-04,
                -3.50790157e-04,
                1.09470845e-03,
                -3.40788020e-03,
                1.02399066e-02,
                -2.63221730e-02,
                5.43868914e-02,
            ],
        ]
    )

    # Computed
    LOO_mean, LOO_cov, LOO_residuals_transformed = loo(K, Y_train)

    # Ensure return values are not None
    assert LOO_mean is not None
    assert LOO_cov is not None
    assert LOO_residuals_transformed is not None

    # Compare
    assert np.allclose(LOO_mean, LOO_mean_true, atol=1e-3)
    assert np.allclose(LOO_cov, LOO_cov_true)
    assert np.allclose(LOO_residuals_transformed, LOO_residuals_transformed_true, atol=1e-3)


def generate_cv_data(
    N_DIM: int = 3,
    N_TRAIN: int = 100,
    N_DUPLICATE_X: int = 0,
    NUM_FOLDS: int = 8,
    NOISE_VAR: float = 0.0,
    SCRAMBLE: bool = True,
) -> tuple[
    NDArray[np.float64],
    NDArray[np.float64],
    list[list[int]],  # folds indices
    NDArray[np.float64],
    torch.Tensor,  # X_train  shape=(N_TRAIN, N_DIM)
    torch.Tensor,  # Y_train  shape=(N_TRAIN)
]:
    """Generate some cross validation data manually for testing."""
    # Set random seed
    random.seed(42)
    torch.manual_seed(42)
    np.random.seed(42)

    # Will generate data using a zero-mean Matern 5/2 GP with these parameters
    KER_SCALE_TRUE: float = 1.0
    KER_LENGTHSCALE_TRUE: torch.Tensor = torch.ones(N_DIM) * 0.5

    # Generate N_TRAIN training Xs, with N_DUPLICATE_X duplicates
    X_train: torch.Tensor = torch.rand(size=(N_TRAIN - N_DUPLICATE_X, N_DIM))

    if N_DUPLICATE_X > 0:
        X_train = torch.cat(
            [
                X_train,
                X_train[np.random.choice(np.arange(N_TRAIN - N_DUPLICATE_X), N_DUPLICATE_X)],
            ]
        )

    # Define kernel and sample training data

    # Kernel
    ker: gpytorch.kernels.Kernel = gpytorch_kernel_Matern(
        outputscale=KER_SCALE_TRUE,
        lengthscale=KER_LENGTHSCALE_TRUE,
    )

    # Covariance matrix; shape=(N_TRAIN, N_TRAIN)
    K: gpytorch.lazy.LazyEvaluatedKernelTensor = ker(X_train)  # type: ignore

    # Distribution
    normal_rv: gpytorch.distributions.Distribution = gpytorch.distributions.MultivariateNormal(
        mean=torch.zeros(K.shape[0]),  # vector; shape=(N_TRAIN)
        covariance_matrix=K,  # matrix; shape=(N_TRAIN, N_TRAIN)
    )

    # Noise
    noise: float | torch.Tensor = 0.0
    if NOISE_VAR != 0.0:
        _noise_rv = gpytorch.distributions.MultivariateNormal(
            mean=torch.zeros(N_TRAIN),
            covariance_matrix=torch.eye(N_TRAIN) * NOISE_VAR,
        )
        noise = _noise_rv.sample()

    # Sample training data
    Y_train: torch.Tensor = normal_rv.sample() + noise

    # Create a list of index subsets
    FOLDS_INDICES: list[list[int]]
    if NUM_FOLDS == N_TRAIN:
        FOLDS_INDICES = [[i] for i in range(N_TRAIN)]
    else:
        # Note: The following sampling approach will not work if NUM_FOLDS is
        # very big (wrt N_TRAIN), but we will only use it for some examples
        # where NUM_FOLDS << N_TRAIN

        folds_end: NDArray[np.float64]
        folds_end = np.random.multinomial(
            N_TRAIN,
            np.ones(NUM_FOLDS) / NUM_FOLDS,
            size=1,
        )[0].cumsum()  # last index of each fold

        folds_startstop: NDArray[np.int_]
        folds_startstop = np.insert(folds_end, 0, 0, axis=0)

        FOLDS_INDICES = [list(range(folds_startstop[i], folds_startstop[i + 1])) for i in range(NUM_FOLDS)]

    if SCRAMBLE:
        _rnd_idx = np.random.permutation(N_TRAIN)  # Randomized indices
        FOLDS_INDICES = [list(_rnd_idx[idx]) for idx in FOLDS_INDICES]

    check_folds_indices(FOLDS_INDICES, N_TRAIN)

    # Fit GP: Create and fit GP model to training data
    _mean: gpytorch.means.Mean = gpytorch_mean_constant(
        0.0,
        fixed=True,
    )
    _likelihood: gpytorch.likelihoods.Likelihood = gpytorch_likelihood_gaussian(
        variance=max(1e-6, NOISE_VAR),
        fixed=False,
    )
    model: ExactGPModel = ExactGPModel(
        X_train,
        Y_train,  # Training data
        _mean,  # Mean function
        ker,  # Kernel
        _likelihood,  # Likelihood
        "",  # Name and path for save/load
        "",
    )

    # Run CV manually
    model.eval_mode()
    # Collect residual means and variances of all folds
    # (residual = observed - predicted)
    _residual_means: list[NDArray[np.float64]] = []
    _residual_vars: list[NDArray[np.float64]] = []
    for i in range(NUM_FOLDS):
        # Split on i-th fold
        fold_X_test, fold_X_train = split_test_train_fold(FOLDS_INDICES, X_train, i)
        fold_Y_test, fold_Y_train = split_test_train_fold(FOLDS_INDICES, Y_train, i)
        # Set training data
        model.set_train_data(inputs=fold_X_train, targets=fold_Y_train, strict=False)
        # Predict on test data
        m, v = model.predict(fold_X_test, latent=False)
        _residual_means.append((fold_Y_test - m).numpy())
        _residual_vars.append(v.numpy())

    # Concatenate and sort so that the residuals correspond to observation 1,2,3 etc.
    cv_residual_means: NDArray[np.float64]
    cv_residual_vars: NDArray[np.float64]
    cv_residual_means = np.block(_residual_means).flatten()
    cv_residual_vars = np.block(_residual_vars).flatten()

    folds_concat = functools.reduce(operator.iadd, FOLDS_INDICES, [])
    idx_sort = list(np.argsort(folds_concat))
    cv_residual_means = cv_residual_means[idx_sort]
    cv_residual_vars = cv_residual_vars[idx_sort]

    return (
        cv_residual_means,
        cv_residual_vars,
        FOLDS_INDICES,
        K.to_dense().detach().numpy(),
        X_train,
        Y_train,
    )


def multitest_loo(
    N_DIM: int,
    N_TRAIN: int,
    NOISE_VAR: float,
    N_DUPLICATE_X: int,
) -> None:
    """Used for running multiple tests of loo().

    Checks that mean and variance are the same as if the residuals were computed
    in a loop.
    This does NOT check covariance and transformed residuals
    """
    # Generate residuals
    (
        cv_residual_means,
        cv_residual_vars,
        folds,
        K,
        X_train,
        Y_train,
    ) = generate_cv_data(
        N_DIM,
        N_TRAIN,
        N_DUPLICATE_X,
        N_TRAIN,
        NOISE_VAR,
        SCRAMBLE=False,
    )

    # Compute residuals from cholesky factor incl jitter
    LOO_mean, LOO_cov, LOO_residuals_transformed = loo(
        K,
        Y_train.numpy(),
        noise_variance=max(1e-6, NOISE_VAR),
    )
    assert LOO_mean is not None
    assert LOO_cov is not None
    LOO_var = LOO_cov.diagonal()

    # Check
    assert np.allclose(LOO_var, cv_residual_vars, atol=1e-3)
    assert np.allclose(LOO_mean, cv_residual_means, atol=1e-3)


def test_loo_noiseless() -> None:
    """Test that loo formula gives same result as computing the residuals in a loop.

    No observational noise, no duplicates
    """
    multitest_loo(N_DIM=3, N_TRAIN=50, NOISE_VAR=0, N_DUPLICATE_X=0)


def test_loo_noise() -> None:
    """Test that loo formula gives same result as computing the residuals in a loop.

    With observational noise, no duplicates
    """
    multitest_loo(N_DIM=3, N_TRAIN=50, NOISE_VAR=0.3, N_DUPLICATE_X=0)


def test_loo_noise_dupl() -> None:
    """Test that loo formula gives same result as computing the residuals in a loop.

    With observational noise, with duplicates
    """
    multitest_loo(N_DIM=3, N_TRAIN=50, NOISE_VAR=0.3, N_DUPLICATE_X=30)


def multitest_multifold(
    N_DIM: int,
    N_TRAIN: int,
    NUM_FOLDS: int,
    NOISE_VAR: float,
    N_DUPLICATE_X: int,
    SCRAMBLE: bool,
) -> None:
    """Used for running multiple tests of multifold().

    Checks that mean and variance are the same as if the residuals were computed
    in a loop.
    This does NOT check covariance and transformed residuals
    """
    # Generate residuals
    (
        cv_residual_means,
        cv_residual_vars,
        folds,
        K,
        X_train,
        Y_train,
    ) = generate_cv_data(
        N_DIM,
        N_TRAIN,
        N_DUPLICATE_X,
        NUM_FOLDS,
        NOISE_VAR,
        SCRAMBLE,
    )

    # Compute residuals from cholesky factor incl jitter
    CV_mean, CV_cov, CV_residuals_transformed = multifold(
        K,
        Y_train.numpy(),
        folds,
        noise_variance=max(1e-6, NOISE_VAR),
    )
    assert CV_mean is not None
    assert CV_cov is not None
    CV_var = CV_cov.diagonal()

    # Check
    assert np.allclose(CV_var, cv_residual_vars, atol=1e-4)
    assert np.allclose(CV_mean, cv_residual_means, atol=1e-3)


def test_multifold_noiseless() -> None:
    """Test that multifold formula gives same result as computing the residuals in a loop.

    No observational noise, no duplicates
    """
    multitest_multifold(
        N_DIM=3,
        N_TRAIN=100,
        NUM_FOLDS=8,
        NOISE_VAR=0,
        N_DUPLICATE_X=0,
        SCRAMBLE=True,
    )


def test_multifold_noise() -> None:
    """Test that multifold formula gives same result as computing the residuals in a loop.

    With observational noise, no duplicates
    """
    multitest_multifold(
        N_DIM=3,
        N_TRAIN=100,
        NUM_FOLDS=8,
        NOISE_VAR=0.3,
        N_DUPLICATE_X=0,
        SCRAMBLE=True,
    )


def test_multifold_noise_dupl() -> None:
    """Test that multifold formula gives same result as computing the residuals in a loop.

    With observational noise, with duplicates
    """
    multitest_multifold(
        N_DIM=3,
        N_TRAIN=100,
        NUM_FOLDS=8,
        NOISE_VAR=0.3,
        N_DUPLICATE_X=30,
        SCRAMBLE=True,
    )


def test_loo_multifold() -> None:
    """Test that LOO and multifold CV gives the same results when fold size = 1."""
    # Generate some data
    N = 100
    NOISE_VAR = 0.23
    (
        cv_residual_means,
        cv_residual_vars,
        folds,
        K,
        X_train,
        Y_train,
    ) = generate_cv_data(
        N_DIM=2,
        N_TRAIN=N,
        N_DUPLICATE_X=20,
        NUM_FOLDS=N,
        NOISE_VAR=NOISE_VAR,
        SCRAMBLE=False,
    )

    CV_mean, CV_cov, CV_residuals_transformed = multifold(
        K,
        Y_train.numpy(),
        folds,
        noise_variance=max(1e-6, NOISE_VAR),
    )
    assert CV_mean is not None
    assert CV_cov is not None
    assert CV_residuals_transformed is not None

    LOO_mean, LOO_cov, LOO_residuals_transformed = loo(
        K,
        Y_train.numpy(),
        noise_variance=max(1e-6, NOISE_VAR),
    )
    assert LOO_mean is not None
    assert LOO_cov is not None
    assert LOO_residuals_transformed is not None

    # Compute multifold CV and LOO
    assert np.allclose(CV_mean, LOO_mean)
    assert np.allclose(CV_cov, LOO_cov)
    assert np.allclose(CV_residuals_transformed, LOO_residuals_transformed)
````

## File: tests/test_metrics.py
````python
"""Unit tests for metrics.py, ensuring log-prob calculations match scipy.stats and main evaluation methods."""

import numpy as np
from scipy import stats

from gp_diagnostics.metrics import log_prob_normal, log_prob_standard_normal


def test_log_prob_normal() -> None:
    """Test that log_prob_normal matches scipy.stats for random data."""
    # Generate some data
    rng = np.random.default_rng(12)
    N = 17

    C = rng.random(size=(N, N))
    C = C.dot(C.T)
    Y = rng.multivariate_normal(mean=np.zeros(N), cov=C)

    # Compute likelihood
    L = np.linalg.cholesky(C)
    loglik = log_prob_normal(L, Y)
    loglik_scipy = stats.multivariate_normal.logpdf(Y, mean=np.zeros(N), cov=C)

    np.testing.assert_allclose(loglik, loglik_scipy)


def test_log_prob_standard_normal() -> None:
    """Test that log_prob_standard_normal matches scipy.stats for standard normal data."""
    # Generate some data
    rng = np.random.default_rng(12)
    N = 16

    Y = rng.multivariate_normal(mean=np.zeros(N), cov=np.eye(N))

    # Compute likelihood
    loglik = log_prob_standard_normal(Y)
    loglik_scipy = stats.multivariate_normal.logpdf(Y, mean=np.zeros(N), cov=np.eye(N))

    np.testing.assert_allclose(loglik, loglik_scipy)
````

## File: tests/test_utils_checks.py
````python
"""Unit tests for checks.py utility functions (numeric array, square, lower triangular)."""

import numpy as np

from gp_diagnostics.utils.checks import is_lower_triang, is_numeric_np_array, is_square


def test_is_numeric_np_array() -> None:
    """Test is_numeric_np_array with various valid and invalid inputs."""
    # These should be ok
    assert is_numeric_np_array(np.array([1, 2, 2.3]))
    assert is_numeric_np_array(np.array([[1, 2, 2.3], [0, 1e-6, 0.001]]))
    assert is_numeric_np_array(np.array(200))

    # These should not be ok
    assert not is_numeric_np_array(np.array([[1, 2, 2.3], [0, 0.001]], dtype=object))
    assert not is_numeric_np_array(np.array(None))
    assert not is_numeric_np_array(np.array([[1, 2, 2.3], [0, 0.001, "a"]]))
    assert not is_numeric_np_array("a")
    assert not is_numeric_np_array(50)
    assert not is_numeric_np_array([1, 2, 3])


def test_is_square() -> None:
    """Test is_square with squares, non-squares, etc."""
    # These should be ok
    assert is_square(np.ones((1, 1)))
    assert is_square(np.ones((14, 14)))

    # These should not be ok
    assert not is_square(np.ones((13, 14)))
    assert not is_square(np.ones((3, 2)))
    assert not is_square(np.ones((3, 3, 3)))
    assert not is_square(np.array(2))
    assert not is_square(np.array([1, 2, 3]))


def test_is_lower_triang() -> None:
    """Test is_lower_triang with valid and invalid matrices."""
    # These should be ok
    assert is_lower_triang(np.array([[0.2, 0, 0], [3, 2.2, 0], [1, 2, 4]]))
    assert is_lower_triang(np.array([[1, 0], [2, 2.2]]))
    assert is_lower_triang(np.array([[1]]))

    # These should not be ok
    assert not is_lower_triang(np.array([[1, 2, 2.3], [0, 2.2, 3], [0.1, 0, 4]]))
````

## File: tests/test_utils_linalg.py
````python
"""Unit tests for linalg.py, verifying triangular solves, Cholesky-based inversion, and trace product operations."""

import numpy as np

from gp_diagnostics.utils.linalg import (
    chol_inv,
    mulinv_solve,
    mulinv_solve_rev,
    symmetrify,
    traceprod,
    triang_solve,
)


def random_array(N: int, seed: int) -> np.ndarray:
    """Returns a random N-dimensional array with a given seed."""
    rng = np.random.default_rng(seed)
    return rng.random(size=N)


def random_matrix(N: int, M: int, seed: int) -> np.ndarray:
    """Returns a random NxM matrix with a given seed."""
    rng = np.random.default_rng(seed)
    return rng.random(size=(N, M))


def random_lower_triang_matrix(N: int, seed: int) -> np.ndarray:
    """Returns a random NxN lower triangular matrix by Cholesky of a random SPD matrix."""
    tmp = random_matrix(N, N, seed)
    return np.linalg.cholesky(tmp.dot(tmp.T))


def test_symmetrify() -> None:
    """Tests symmetrify in-place for random lower triangular matrices."""
    N = 10

    A = random_lower_triang_matrix(N, 77)
    symmetrify(A, upper=False)

    assert np.allclose(A - A.T, np.zeros((N, N)))

    A = random_lower_triang_matrix(N, 77)
    symmetrify(A.T, upper=True)

    assert np.allclose(A - A.T, np.zeros((N, N)))


def test_triang_solve() -> None:
    """Tests direct and transposed triangular solves with random data."""
    N = 10
    M = 4

    # Generate matrices
    L = random_lower_triang_matrix(N, 42)
    B = random_matrix(N, M, 43)

    # Solve L*X = B and check
    X = triang_solve(L, B, lower=True, trans=False)
    assert X.shape == (N, M)
    assert np.allclose(B, L.dot(X))

    # Solve L.T*X = B and check
    X = triang_solve(L.T, B, lower=False, trans=False)
    assert X.shape == (N, M)
    assert np.allclose(B, L.T.dot(X))

    # Solve L.T*X = B and check
    X = triang_solve(L, B, lower=True, trans=True)
    assert X.shape == (N, M)
    assert np.allclose(B, L.T.dot(X))

    # Solve L*X = B and check
    X = triang_solve(L.T, B, lower=False, trans=True)
    assert X.shape == (N, M)
    assert np.allclose(B, L.dot(X))


def test_triang_solve_with_B_being_1D_array() -> None:
    """Tests direct and transposed triangular solves with a 1D RHS array and random data."""
    N = 10

    # Generate matrices
    L = random_lower_triang_matrix(N, 42)
    B = random_array(N, 43)

    # Solve L*X = B and check
    X = triang_solve(L, B, lower=True, trans=False)
    assert X.shape == (N,)
    assert np.allclose(B, L.dot(X))

    # Solve L.T*X = B and check
    X = triang_solve(L.T, B, lower=False, trans=False)
    assert X.shape == (N,)
    assert np.allclose(B, L.T.dot(X))

    # Solve L.T*X = B and check
    X = triang_solve(L, B, lower=True, trans=True)
    assert X.shape == (N,)
    assert np.allclose(B, L.T.dot(X))

    # Solve L*X = B and check
    X = triang_solve(L.T, B, lower=False, trans=True)
    assert X.shape == (N,)
    assert np.allclose(B, L.dot(X))


def test_mulinv_solve() -> None:
    """Tests mulinv_solve for A=F*F^T with random data."""
    N = 10
    M = 4

    # Generate matrices
    L = random_lower_triang_matrix(N, 42)
    B = random_matrix(N, M, 43)
    A = L.dot(L.T)

    # Solve A*X = B and check
    X = mulinv_solve(L, B, lower=True)
    assert X.shape == (N, M)
    assert np.allclose(B, A.dot(X))

    # Solve A*X = B and check
    X = mulinv_solve(L.T, B, lower=False)
    assert X.shape == (N, M)
    assert np.allclose(B, L.T.dot(L).dot(X))


def test_mulinv_solve_with_B_being_1D_array() -> None:
    """Tests mulinv_solve for A=F*F^T with a 1D RHS array and random data."""
    N = 10

    # Generate matrix and array
    L = random_lower_triang_matrix(N, 42)
    B = random_array(N, 43)
    A = L.dot(L.T)

    # Solve A*X = B and check
    X = mulinv_solve(L, B, lower=True)
    assert X.shape == (N,)
    assert np.allclose(B, A.dot(X))

    # Solve A*X = B and check
    X = mulinv_solve(L.T, B, lower=False)
    assert X.shape == (N,)
    assert np.allclose(B, L.T.dot(L).dot(X))


def test_mulinv_solve_rev() -> None:
    """Tests the reversed (X*A=B) version of mulinv_solve with random data."""
    N = 10
    M = 4

    # Generate matrices
    L = random_lower_triang_matrix(N, 42)
    B = random_matrix(M, N, 43)
    A = L.dot(L.T)

    # Solve X*A = B and check
    X = mulinv_solve_rev(L, B, lower=True)
    assert X.shape == (M, N)
    assert np.allclose(B, X.dot(A))

    # Solve X*A = B and check
    X = mulinv_solve_rev(L.T, B, lower=False)
    assert X.shape == (M, N)
    assert np.allclose(B, X.dot(L.T.dot(L)))


def test_mulinv_solve_rev_with_B_being_1D_array() -> None:
    """Tests the reversed (X*A=B) version of mulinv_solve for A=F*F^T with a 1D RHS array and random data."""
    N = 10

    # Generate matrices
    L = random_lower_triang_matrix(N, 42)
    B = random_array(N, 43)
    A = L.dot(L.T)

    # Solve X*A = B and check
    X = mulinv_solve_rev(L, B, lower=True)
    assert X.shape == (N,)
    assert np.allclose(B, X.dot(A))

    # Solve X*A = B and check
    X = mulinv_solve_rev(L.T, B, lower=False)
    assert X.shape == (N,)
    assert np.allclose(B, X.dot(L.T.dot(L)))


def test_chol_inv() -> None:
    """Tests chol_inv for correctness vs. np.linalg.inv using random data."""
    N = 10

    # Generate matrices
    L = random_lower_triang_matrix(N, 42)
    A = L.dot(L.T)

    # Invert and check
    A_inv_true = np.linalg.inv(A)
    A_inv = chol_inv(L)

    assert np.allclose(A_inv_true, A_inv)


def test_traceprod() -> None:
    """Tests traceprod vs. direct diagonal sum for random data."""
    N = 10
    M = 8

    # Generate matrices
    A = random_matrix(N, M, 42)
    B = random_matrix(M, N, 43)

    # Compute and check
    trace = traceprod(A, B)
    trace_true = A.dot(B).diagonal().sum()
    assert np.allclose(trace, trace_true)
````

## File: tests/test_utils_stats.py
````python
"""Unit tests for stats.py, verifying normal QQ computations and split_test_train_fold functionality."""

import numpy as np

from gp_diagnostics.utils.stats import snorm_qq, split_test_train_fold


def test_snorm_qq_equalR() -> None:
    """Check that snorm_qq() matches known R's extRemes.qqnorm output for standard normal sample."""
    # Inputs (generated from a standard normal variable)
    x = np.array(
        [
            -0.325459352843027,
            -0.894532916553418,
            -0.651097296010105,
            0.431962456402265,
            1.11307123980658,
            0.16432330947537,
            -1.3009717579692,
            -1.20616922164046,
            -0.862389244884325,
            -0.783944517858679,
            -0.86289300469735,
            -0.452842255748436,
            -1.32372331616317,
            -1.21546884455723,
            -0.913805791530649,
            0.412898096437753,
            -0.044206507445586,
            -0.106133827148484,
            -1.21579545950289,
            -0.791685174905493,
        ]
    )

    # Output
    q_sample, q_snorm, q_snorm_upper, q_snorm_lower = snorm_qq(x)

    # Output computed from extRemes.qqnorm in R
    R_lower = np.array(
        [
            np.nan,
            np.nan,
            np.nan,
            np.nan,
            -1.8436377834902,
            -1.38767528310563,
            -1.11410079939012,
            -0.905433177563493,
            -0.730252354751578,
            -0.575081844939952,
            -0.432695573891231,
            -0.298612213033157,
            -0.169712878082784,
            -0.0435840752400704,
            0.0818524407565779,
            0.208595460977847,
            0.338795151235563,
            0.475032155363908,
            0.620750108775009,
            0.781065359836158,
        ]
    )

    R_upper = np.array(
        [
            -0.781065359836158,
            -0.620750108775009,
            -0.475032155363908,
            -0.338795151235563,
            -0.208595460977847,
            -0.0818524407565779,
            0.0435840752400707,
            0.169712878082784,
            0.298612213033157,
            0.432695573891231,
            0.575081844939952,
            0.730252354751577,
            0.905433177563493,
            1.11410079939012,
            1.38767528310563,
            1.8436377834902,
            np.nan,
            np.nan,
            np.nan,
            np.nan,
        ]
    )

    R_snorm = np.array(
        [
            -1.95996398454005,
            -1.43953147093846,
            -1.15034938037601,
            -0.93458929107348,
            -0.755415026360469,
            -0.597760126042478,
            -0.453762190169879,
            -0.318639363964375,
            -0.189118426272792,
            -0.0627067779432138,
            0.0627067779432138,
            0.189118426272792,
            0.318639363964375,
            0.45376219016988,
            0.597760126042478,
            0.755415026360469,
            0.93458929107348,
            1.15034938037601,
            1.43953147093846,
            1.95996398454005,
        ]
    )

    R_data = np.array(
        [
            -1.32372331616317,
            -1.3009717579692,
            -1.21579545950289,
            -1.21546884455723,
            -1.20616922164046,
            -0.913805791530649,
            -0.894532916553418,
            -0.86289300469735,
            -0.862389244884325,
            -0.791685174905493,
            -0.783944517858679,
            -0.651097296010105,
            -0.452842255748436,
            -0.325459352843027,
            -0.106133827148484,
            -0.044206507445586,
            0.16432330947537,
            0.412898096437753,
            0.431962456402265,
            1.11307123980658,
        ]
    )

    # Check expected types
    assert isinstance(x, np.ndarray)
    assert isinstance(q_sample, np.ndarray)
    assert isinstance(q_snorm, np.ndarray)
    assert isinstance(q_snorm_upper, np.ndarray)
    assert isinstance(q_snorm_lower, np.ndarray)

    # Check expected shapes
    N: int = 20
    assert x.shape == (N,)
    assert q_sample.shape == (N,)
    assert q_snorm.shape == (N,)
    assert q_snorm_upper.shape == (N,)
    assert q_snorm_lower.shape == (N,)

    # Check that results match the reference output computed from extRemes.qqnorm in R
    assert np.allclose(q_snorm_lower, R_lower, equal_nan=True)
    assert np.allclose(q_snorm_upper, R_upper, equal_nan=True)
    assert np.allclose(q_snorm, R_snorm)
    assert np.allclose(q_sample, R_data)


def test_split_test_train_fold() -> None:
    """Check correctness of splitting for test vs. train folds."""
    folds = [[1, 4, 5], [0], [7], [2, 3, 6], [9, 8]]

    x = np.arange(10, dtype=np.float64) * 1.1

    y1 = np.array([1.1, 4.4, 5.5, 0.0, 7.7, 2.2, 3.3, 6.6])
    y2 = np.array([9.9, 8.8])
    x_test, x_train = split_test_train_fold(folds, x, 4)
    assert np.allclose(y1, x_train)
    assert np.allclose(y2, x_test)

    y1 = np.array([1.1, 4.4, 5.5, 7.7, 2.2, 3.3, 6.6, 9.9, 8.8])
    y2 = np.array([0.0])
    x_test, x_train = split_test_train_fold(folds, x, 1)
    assert np.allclose(y1, x_train)
    assert np.allclose(y2, x_test)

    x_test, x_train = split_test_train_fold(folds, x.reshape(-1, 1), 1)
    assert np.allclose(y1.reshape(-1, 1), x_train)
    assert np.allclose(y2.reshape(-1, 1), x_test)
````

## File: tests/utils.py
````python
"""Helper classes and functions for testing, including a simple ExactGPModel and GPyTorch utilities."""

import contextlib
from typing import TYPE_CHECKING, Any

import gpytorch
import gpytorch.constraints
import torch

if TYPE_CHECKING:
    from numpy.typing import NDArray


class ExactGPModel(gpytorch.models.ExactGP):
    """Model for standard GP regression."""

    def __init__(
        self,
        train_x: torch.Tensor,
        train_y: torch.Tensor,
        mean_module: gpytorch.means.Mean,
        covar_module: gpytorch.kernels.Kernel,
        likelihood: gpytorch.likelihoods.Likelihood,
        path: str = "",
        name: str = "",
    ) -> None:
        # Note: Overwriting the declaration of self.likelihood is necessary to
        # make explicit to code linters that likelihood is not optional in our
        # implementation, i.e. likelihood cannot be None. (This is different
        # from the ExactGP base class implementation where likelihood can also
        # be None.)
        self.likelihood: gpytorch.likelihoods.Likelihood

        super().__init__(train_x, train_y, likelihood)
        assert self.likelihood is not None

        # For saving and loading
        self.path = path
        self.name = name
        self.param_fname = self.path + self.name + ".pth"

        # Mean and covariance functions
        self.mean_module = mean_module
        self.covar_module = covar_module

    def forward(self, x: torch.Tensor) -> gpytorch.distributions.MultivariateNormal:
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)

    def eval_mode(self) -> None:
        """Set model in evaluation mode."""
        self.eval()
        self.likelihood.eval()

    def train_mode(self) -> None:
        """Set in training mode."""
        self.train()
        self.likelihood.train()

    def predict(
        self,
        x: torch.Tensor,
        latent: bool = True,
        CG_tol: float = 0.1,
        full_cov: bool = False,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Return mean and covariance at x.

        Input:
        x         -      tensor of size dim * N containing N inputs
        latent    -      latent = True ->  using latent GP
                         latent = False -> using observed GP (incl. likelihood)
        CG_tol    -      Conjugate Gradient tolerance for evaluation
        full_cov  -      full_cov = False -> Return only diagonal (variances)

        Output:
        mean and covariance
        """
        mean: torch.Tensor
        var: torch.Tensor
        with torch.no_grad(), gpytorch.settings.eval_cg_tolerance(CG_tol):
            # Latent distribution
            dist: torch.distributions.Distribution = self.__call__(x)

            # Observational distribution
            if not latent:
                _dist = self.likelihood(dist)
                if isinstance(_dist, torch.distributions.Distribution):
                    dist = _dist

            # Extract mean and covariance
            assert isinstance(dist, gpytorch.distributions.MultivariateNormal)
            # if isinstance(dist, gpytorch.distributions.MultivariateNormal):
            mean = dist.mean.cpu()
            var = (
                dist.covariance_matrix.cpu()  # type: ignore
                if full_cov
                else dist.variance.cpu()
            )

        return mean, var

    def print_parameters(self) -> None:
        """Print actual (not raw) parameters."""
        _constant_mean: torch.Tensor | str = "--"
        with contextlib.suppress(Exception):
            _constant_mean = self.mean_module.constant.item()  # type: ignore

        _noise: torch.Tensor | str = "--"
        with contextlib.suppress(Exception):
            _noise = self.likelihood.noise_covar.noise.item()  # type: ignore

        _lengthscale: NDArray[Any] | str = "--"
        with contextlib.suppress(Exception):
            _lengthscale = self.covar_module.base_kernel.lengthscale.detach().numpy()[0]

        _outputscale: torch.Tensor | str = "--"
        with contextlib.suppress(Exception):
            _outputscale = self.covar_module.outputscale.item()  # type: ignore

    def save(self) -> None:
        """Save GP model parameters to self.path."""
        torch.save(self.state_dict(), self.param_fname)

    def load(self) -> None:
        """Load GP model parameters from self.path."""
        self.load_state_dict(torch.load(self.param_fname))


def gpytorch_kernel_Matern(
    outputscale: float,
    lengthscale: torch.Tensor,
    nu: float = 2.5,
    lengthscale_constraint: gpytorch.constraints.Interval | None = None,
) -> gpytorch.kernels.Kernel:
    """Return a scaled Matern kernel with specified output scale and lengthscale."""
    lengthscale_constraint = lengthscale_constraint or gpytorch.constraints.Positive()
    ker_mat = gpytorch.kernels.MaternKernel(
        nu=nu,
        ard_num_dims=len(lengthscale),
        lengthscale_constraint=lengthscale_constraint,
    )
    ker_mat.lengthscale = lengthscale
    ker = gpytorch.kernels.ScaleKernel(ker_mat)
    ker.outputscale = outputscale

    return ker


def gpytorch_mean_constant(val: float, fixed: bool = True) -> gpytorch.means.Mean:
    """Return a constant mean function.

    fixed = True -> Do not update mean function during training
    """
    mean = gpytorch.means.ConstantMean()
    mean.initialize(constant=val)
    assert isinstance(mean.constant, torch.Tensor)
    mean.constant.requires_grad = not fixed

    return mean


def gpytorch_likelihood_gaussian(
    variance: float, variance_lb: float = 1e-6, fixed: bool = True
) -> gpytorch.likelihoods.Likelihood:
    """Return a Gaussian likelihood.

    fixed = True -> Do not update during training
    variance_lb = lower bound
    """
    likelihood = gpytorch.likelihoods.GaussianLikelihood(noise_constraint=gpytorch.constraints.GreaterThan(variance_lb))
    likelihood.initialize(noise=variance)
    # @TODO: Base type of likelihood is gpytorch.Module, not torch.Tensor .
    #   Natively, hence, likelihood does not have an attribute
    #   'requires_grad'.
    #   What the following code effectively does is to dynamically
    #   add an attribute with name='requires_grad' to the likelihood instance
    #   and assign it a boolean value.
    #   @AGRE / @ELD: Is this really what you intended, and is it necessary?
    #   CLAROS, 2022-11-01
    likelihood.requires_grad = not fixed

    return likelihood
````

## File: tox.toml
````toml
requires = [
    "tox>=4.24.1",
    "tox-uv>=1.20.1",
]
env_list = [
    "3.10",
    "3.11",
    "3.12",
]

[env_run_base]
description = "run unit tests"
runner = "uv-venv-lock-runner"
with_dev = true
commands = [
    ["pytest"],
]
````
